@@ -5,21 +5,55 @@ natsort public API.
 
 The majority of the "work" is defined in utils.py.
 """
-from __future__ import absolute_import, division, print_function, unicode_literals
 
-import sys
+import platform
 from functools import partial
 from operator import itemgetter
-from warnings import warn
+from typing import (
+    Any,
+    Callable,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+    cast,
+    overload,
+)
 
 import natsort.compat.locale
 from natsort import utils
-from natsort.compat.py23 import py23_cmp, py23_str, u_format
-from natsort.ns_enum import ns, ns_DUMB
+from natsort.ns_enum import NSType, NS_DUMB, ns
+from natsort.utils import (
+    KeyType,
+    MaybeKeyType,
+    NatsortInType,
+    NatsortOutType,
+    StrBytesNum,
+    StrBytesPathNum,
+)
+
+# Common input and output types
+Iter_ns = Iterable[NatsortInType]
+Iter_any = Iterable[Any]
+List_ns = List[NatsortInType]
+List_any = List[Any]
+List_int = List[int]
+
+# The type that natsort_key returns
+NatsortKeyType = Callable[[NatsortInType], NatsortOutType]
+
+# Types for os_sorted
+OSSortInType = Iterable[Optional[StrBytesPathNum]]
+OSSortOutType = Tuple[Union[StrBytesNum, Tuple[StrBytesNum, ...]], ...]
+OSSortKeyType = Callable[[Optional[StrBytesPathNum]], OSSortOutType]
+Iter_path = Iterable[Optional[StrBytesPathNum]]
+List_path = List[StrBytesPathNum]
 
 
-@u_format
-def decoder(encoding):
+def decoder(encoding: str) -> Callable[[NatsortInType], NatsortInType]:
     """
     Return a function that can be used to decode bytes to unicode.
 
@@ -60,8 +94,7 @@ def decoder(encoding):
     return partial(utils.do_decoding, encoding=encoding)
 
 
-@u_format
-def as_ascii(s):
+def as_ascii(s: NatsortInType) -> NatsortInType:
     """
     Function to decode an input with the ASCII codec, or return as-is.
 
@@ -84,8 +117,7 @@ def as_ascii(s):
     return utils.do_decoding(s, "ascii")
 
 
-@u_format
-def as_utf8(s):
+def as_utf8(s: NatsortInType) -> NatsortInType:
     """
     Function to decode an input with the UTF-8 codec, or return as-is.
 
@@ -108,8 +140,9 @@ def as_utf8(s):
     return utils.do_decoding(s, "utf-8")
 
 
-@u_format
-def natsort_keygen(key=None, alg=ns.DEFAULT, **_kwargs):
+def natsort_keygen(
+    key: MaybeKeyType = None, alg: NSType = ns.DEFAULT
+) -> NatsortKeyType:
     """
     Generate a key to sort strings and numbers naturally.
 
@@ -152,19 +185,18 @@ def natsort_keygen(key=None, alg=ns.DEFAULT, **_kwargs):
         >>> a = ['num5.10', 'num-3', 'num5.3', 'num2']
         >>> a.sort(key=natsort_keygen(alg=ns.REAL))
         >>> a
-        [{u}'num-3', {u}'num2', {u}'num5.10', {u}'num5.3']
+        ['num-3', 'num2', 'num5.10', 'num5.3']
 
     """
-    # Transform old arguments to the ns enum.
     try:
-        alg = utils.args_to_enum(**_kwargs) | alg
+        ns.DEFAULT | alg
     except TypeError:
         msg = "natsort_keygen: 'alg' argument must be from the enum 'ns'"
-        raise ValueError(msg + ", got {0}".format(py23_str(alg)))
+        raise ValueError(msg + ", got {}".format(str(alg)))
 
-    # Add the _DUMB option if the locale library is broken.
+    # Add the NS_DUMB option if the locale library is broken.
     if alg & ns.LOCALEALPHA and natsort.compat.locale.dumb_sort():
-        alg |= ns_DUMB
+        alg |= NS_DUMB
 
     # Set some variables that will be passed to the factory functions
     if alg & ns.NUMAFTER:
@@ -193,7 +225,7 @@ def natsort_keygen(key=None, alg=ns.DEFAULT, **_kwargs):
     if alg & ns.PATH:
         string_func = utils.parse_path_factory(string_func)
     bytes_func = utils.parse_bytes_factory(alg)
-    num_func = utils.parse_number_factory(alg, sep, pre_sep)
+    num_func = utils.parse_number_or_none_factory(alg, sep, pre_sep)
 
     # Return the natsort key with the parsing path pre-chosen.
     return partial(
@@ -220,8 +252,26 @@ natsort_keygen
 """
 
 
-@u_format
-def natsorted(seq, key=None, reverse=False, alg=ns.DEFAULT, **_kwargs):
+@overload
+def natsorted(
+    seq: Iter_ns, key: None = None, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_ns:
+    ...
+
+
+@overload
+def natsorted(
+    seq: Iter_any, key: KeyType, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_any:
+    ...
+
+
+def natsorted(
+    seq: Iter_any,
+    key: MaybeKeyType = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List_any:
     """
     Sorts an iterable naturally.
 
@@ -255,6 +305,7 @@ def natsorted(seq, key=None, reverse=False, alg=ns.DEFAULT, **_kwargs):
     realsorted : A wrapper for ``natsorted(seq, alg=ns.REAL)``.
     humansorted : A wrapper for ``natsorted(seq, alg=ns.LOCALE)``.
     index_natsorted : Returns the sorted indexes from `natsorted`.
+    os_sorted : Sort according to your operating system's rules.
 
     Examples
     --------
@@ -262,34 +313,32 @@ def natsorted(seq, key=None, reverse=False, alg=ns.DEFAULT, **_kwargs):
 
         >>> a = ['num3', 'num5', 'num2']
         >>> natsorted(a)
-        [{u}'num2', {u}'num3', {u}'num5']
+        ['num2', 'num3', 'num5']
 
     """
-    key = natsort_keygen(key, alg, **_kwargs)
-    return sorted(seq, reverse=reverse, key=key)
+    return sorted(seq, reverse=reverse, key=natsort_keygen(key, alg))
 
 
-@u_format
-def versorted(seq, key=None, reverse=False, alg=ns.DEFAULT, **_kwargs):
-    """
-    Identical to :func:`natsorted`.
-
-    This function is deprecated as of :mod:`natsort` 5.5.0, and will be
-    removed in 6.0.0.
-
-    See Also
-    --------
-    natsorted
-
-    """
-    msg = "versorted is deprecated as of 5.5.0 and will be removed in 6.0.0, "
-    msg += "please use natsorted instead."
-    warn(msg, DeprecationWarning, stacklevel=2)
-    return natsorted(seq, key, reverse, alg, **_kwargs)
+@overload
+def humansorted(
+    seq: Iter_ns, key: None = None, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_ns:
+    ...
 
 
-@u_format
-def humansorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+@overload
+def humansorted(
+    seq: Iter_any, key: KeyType, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_any:
+    ...
+
+
+def humansorted(
+    seq: Iter_any,
+    key: MaybeKeyType = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List_any:
     """
     Convenience function to properly sort non-numeric characters.
 
@@ -333,16 +382,34 @@ def humansorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
 
         >>> a = ['Apple', 'Banana', 'apple', 'banana']
         >>> natsorted(a)
-        [{u}'Apple', {u}'Banana', {u}'apple', {u}'banana']
+        ['Apple', 'Banana', 'apple', 'banana']
         >>> humansorted(a)
-        [{u}'apple', {u}'Apple', {u}'banana', {u}'Banana']
+        ['apple', 'Apple', 'banana', 'Banana']
 
     """
     return natsorted(seq, key, reverse, alg | ns.LOCALE)
 
 
-@u_format
-def realsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+@overload
+def realsorted(
+    seq: Iter_ns, key: None = None, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_ns:
+    ...
+
+
+@overload
+def realsorted(
+    seq: Iter_any, key: KeyType, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_any:
+    ...
+
+
+def realsorted(
+    seq: Iter_any,
+    key: MaybeKeyType = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List_any:
     """
     Convenience function to properly sort signed floats.
 
@@ -387,16 +454,34 @@ def realsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
 
         >>> a = ['num5.10', 'num-3', 'num5.3', 'num2']
         >>> natsorted(a)
-        [{u}'num2', {u}'num5.3', {u}'num5.10', {u}'num-3']
+        ['num2', 'num5.3', 'num5.10', 'num-3']
         >>> realsorted(a)
-        [{u}'num-3', {u}'num2', {u}'num5.10', {u}'num5.3']
+        ['num-3', 'num2', 'num5.10', 'num5.3']
 
     """
     return natsorted(seq, key, reverse, alg | ns.REAL)
 
 
-@u_format
-def index_natsorted(seq, key=None, reverse=False, alg=ns.DEFAULT, **_kwargs):
+@overload
+def index_natsorted(
+    seq: Iter_ns, key: None = None, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_int:
+    ...
+
+
+@overload
+def index_natsorted(
+    seq: Iter_any, key: KeyType, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_int:
+    ...
+
+
+def index_natsorted(
+    seq: Iter_any,
+    key: MaybeKeyType = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List_int:
     """
     Determine the list of the indexes used to sort the input sequence.
 
@@ -447,45 +532,45 @@ def index_natsorted(seq, key=None, reverse=False, alg=ns.DEFAULT, **_kwargs):
         [2, 0, 1]
         >>> # Sort both lists by the sort order of a
         >>> order_by_index(a, index)
-        [{u}'num2', {u}'num3', {u}'num5']
+        ['num2', 'num3', 'num5']
         >>> order_by_index(b, index)
-        [{u}'baz', {u}'foo', {u}'bar']
+        ['baz', 'foo', 'bar']
 
     """
+    newkey: KeyType
     if key is None:
         newkey = itemgetter(1)
     else:
 
-        def newkey(x):
-            return key(itemgetter(1)(x))
+        def newkey(x: Any) -> NatsortInType:
+            return cast(KeyType, key)(itemgetter(1)(x))
 
     # Pair the index and sequence together, then sort by element
-    index_seq_pair = [[x, y] for x, y in enumerate(seq)]
-    index_seq_pair.sort(reverse=reverse, key=natsort_keygen(newkey, alg, **_kwargs))
+    index_seq_pair = [(x, y) for x, y in enumerate(seq)]
+    index_seq_pair.sort(reverse=reverse, key=natsort_keygen(newkey, alg))
     return [x for x, _ in index_seq_pair]
 
 
-@u_format
-def index_versorted(seq, key=None, reverse=False, alg=ns.DEFAULT, **_kwargs):
-    """
-    Identical to :func:`index_natsorted`.
-
-    This function is deprecated as of :mod:`natsort` 5.5.0, and will be
-    removed in 6.0.0.
-
-    See Also
-    --------
-    index_natsorted
-
-    """
-    msg = "index_versorted is deprecated as of 5.5.0 and will be removed in 6.0.0, "
-    msg += "please use index_natsorted instead."
-    warn(msg, DeprecationWarning, stacklevel=2)
-    return index_natsorted(seq, key, reverse, alg, **_kwargs)
+@overload
+def index_humansorted(
+    seq: Iter_ns, key: None = None, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_int:
+    ...
 
 
-@u_format
-def index_humansorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+@overload
+def index_humansorted(
+    seq: Iter_any, key: KeyType, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_int:
+    ...
+
+
+def index_humansorted(
+    seq: Iter_any,
+    key: MaybeKeyType = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List_int:
     """
     This is a wrapper around ``index_natsorted(seq, alg=ns.LOCALE)``.
 
@@ -534,8 +619,26 @@ def index_humansorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
     return index_natsorted(seq, key, reverse, alg | ns.LOCALE)
 
 
-@u_format
-def index_realsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+@overload
+def index_realsorted(
+    seq: Iter_ns, key: None = None, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_int:
+    ...
+
+
+@overload
+def index_realsorted(
+    seq: Iter_any, key: KeyType, reverse: bool = False, alg: NSType = ns.DEFAULT
+) -> List_int:
+    ...
+
+
+def index_realsorted(
+    seq: Iter_any,
+    key: MaybeKeyType = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List_int:
     """
     This is a wrapper around ``index_natsorted(seq, alg=ns.REAL)``.
 
@@ -581,8 +684,9 @@ def index_realsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
 
 
 # noinspection PyShadowingBuiltins,PyUnresolvedReferences
-@u_format
-def order_by_index(seq, index, iter=False):
+def order_by_index(
+    seq: Sequence[Any], index: Iterable[int], iter: bool = False
+) -> Iter_any:
     """
     Order a given sequence by an index sequence.
 
@@ -633,67 +737,194 @@ def order_by_index(seq, index, iter=False):
         [2, 0, 1]
         >>> # Sort both lists by the sort order of a
         >>> order_by_index(a, index)
-        [{u}'num2', {u}'num3', {u}'num5']
+        ['num2', 'num3', 'num5']
         >>> order_by_index(b, index)
-        [{u}'baz', {u}'foo', {u}'bar']
+        ['baz', 'foo', 'bar']
 
     """
     return (seq[i] for i in index) if iter else [seq[i] for i in index]
 
 
-if float(sys.version[:3]) < 3:
-    # pylint: disable=unused-variable
-    # noinspection PyUnresolvedReferences,PyPep8Naming
-    class natcmp(object):  # noqa: N801
-        """
-        Compare two objects using a key and an algorithm.
+def numeric_regex_chooser(alg: NSType) -> str:
+    """
+    Select an appropriate regex for the type of number of interest.
 
-        Parameters
-        ----------
-        x : object
-            First object to compare.
+    Parameters
+    ----------
+    alg : ns enum
+        Used to indicate the regular expression to select.
 
-        y : object
-            Second object to compare.
+    Returns
+    -------
+    regex : str
+        Regular expression string that matches the desired number type.
 
-        alg : ns enum, optional
-            This option is used to control which algorithm `natsort`
-            uses when sorting. For details into these options, please see
-            the :class:`ns` class documentation. The default is `ns.INT`.
+    """
+    # Remove the leading and trailing parens
+    return utils.regex_chooser(alg).pattern[1:-1]
 
-        Returns
-        -------
-        out: int
-            0 if x and y are equal, 1 if x > y, -1 if y > x.
 
-        See Also
-        --------
-        natsort_keygen : Generates a key that makes natural sorting possible.
+def _split_apply(v: Any, key: MaybeKeyType = None) -> Iterator[str]:
+    if key is not None:
+        v = key(v)
+    return utils.path_splitter(str(v))
 
-        Examples
-        --------
-        Use `natcmp` just like the builtin `cmp`::
 
-            >>> one = 1
-            >>> two = 2
-            >>> natcmp(one, two)
-            -1
-        """
+# Choose the implementation based on the host OS
+if platform.system() == "Windows":
 
-        cached_keys = {}
+    from ctypes import wintypes, windll  # type: ignore
+    from functools import cmp_to_key
 
-        def __new__(cls, x, y, alg=ns.DEFAULT, *args, **kwargs):
-            try:
-                alg = utils.args_to_enum(**kwargs) | alg
-            except TypeError:
-                msg = "natsort_keygen: 'alg' argument must be " "from the enum 'ns'"
-                raise ValueError(msg + ", got {0}".format(py23_str(alg)))
+    _windows_sort_cmp = windll.Shlwapi.StrCmpLogicalW
+    _windows_sort_cmp.argtypes = [wintypes.LPWSTR, wintypes.LPWSTR]
+    _windows_sort_cmp.restype = wintypes.INT
+    _winsort_key = cmp_to_key(_windows_sort_cmp)
 
-            # Add the _DUMB option if the locale library is broken.
-            if alg & ns.LOCALEALPHA and natsort.compat.locale.dumb_sort():
-                alg |= ns_DUMB
+    def os_sort_keygen(key: MaybeKeyType = None) -> OSSortKeyType:
+        return cast(
+            OSSortKeyType, lambda x: tuple(map(_winsort_key, _split_apply(x, key)))
+        )
 
-            if alg not in cls.cached_keys:
-                cls.cached_keys[alg] = natsort_keygen(alg=alg)
 
-            return py23_cmp(cls.cached_keys[alg](x), cls.cached_keys[alg](y))
+else:
+
+    # For UNIX-based platforms, ICU performs MUCH better than locale
+    # at replicating the file explorer's sort order. We will use
+    # ICU's ability to do basic natural sorting as it also better
+    # replicates than what natsort does by default.
+    #
+    # However, if the user does not have ICU installed then fall back
+    # on natsort's default handling for paths with locale turned on
+    # which will give good results in most cases (e.g. when there aren't
+    # a bunch of special characters).
+    try:
+        import icu
+
+    except ImportError:
+        # No ICU installed
+        def os_sort_keygen(key: MaybeKeyType = None) -> OSSortKeyType:
+            return cast(
+                OSSortKeyType,
+                natsort_keygen(key=key, alg=ns.LOCALE | ns.PATH | ns.IGNORECASE),
+            )
+
+    else:
+        # ICU installed
+        def os_sort_keygen(key: MaybeKeyType = None) -> OSSortKeyType:
+            loc = natsort.compat.locale.get_icu_locale()
+            collator = icu.Collator.createInstance(loc)
+            collator.setAttribute(
+                icu.UCollAttribute.NUMERIC_COLLATION, icu.UCollAttributeValue.ON
+            )
+            return lambda x: tuple(map(collator.getSortKey, _split_apply(x, key)))
+
+
+os_sort_keygen.__doc__ = """
+Generate a sorting key to replicate your file browser's sort order
+
+See :func:`os_sorted` for description and caveats.
+
+Returns
+-------
+out : function
+    A function that parses input for OS path sorting that is
+    suitable for passing as the `key` argument to functions
+    such as `sorted`.
+
+See Also
+--------
+os_sort_key
+os_sorted
+
+Notes
+-----
+On Windows, this will implicitly coerce all inputs to str before
+collating.
+
+"""
+
+os_sort_key = os_sort_keygen()
+os_sort_key.__doc__ = """
+os_sort_key(val)
+The default key to replicate your file browser's sort order
+
+This is the output of :func:`os_sort_keygen` with default values.
+
+See Also
+--------
+os_sort_keygen
+
+"""
+
+
+@overload
+def os_sorted(seq: Iter_path, key: None = None, reverse: bool = False) -> List_path:
+    ...
+
+
+@overload
+def os_sorted(seq: Iter_any, key: KeyType, reverse: bool = False) -> List_any:
+    ...
+
+
+def os_sorted(
+    seq: Iter_any, key: MaybeKeyType = None, reverse: bool = False
+) -> List_any:
+    """
+    Sort elements in the same order as your operating system's file browser
+
+    .. warning::
+
+        The resulting function will generate results that will be
+        differnt depending on your platform. This is intentional.
+
+    On Windows, this will sort with the same order as Windows Explorer.
+
+    On MacOS/Linux, you will get different results depending on whether
+    or not you have :mod:`pyicu` installed.
+
+    - If you have :mod:`pyicu` installed, you will get results that are
+      the same as (or very close to) the same order as your operating
+      system's file browser.
+    - If you do not have :mod:`pyicu` installed, then this will give
+      the same results as if you used ``ns.LOCALE``, ``ns.PATH``,
+      and ``ns.IGNORECASE`` with :func:`natsorted`. If you do not have
+      special characters this will give correct results, but once
+      special characters are added you should lower your expectations.
+
+    It is *strongly* reccommended to have :mod:`pyicu` installed on
+    MacOS/Linux if you want correct sort results.
+
+    It does *not* take into account if a path is a directory or a file
+    when sorting.
+
+    Parameters
+    ----------
+    seq : iterable
+        The input to sort. Each element must be of type str.
+
+    key : callable, optional
+        A key used to determine how to sort each element of the sequence.
+        It should accept a single argument and return a single value.
+
+    reverse : {{True, False}}, optional
+        Return the list in reversed sorted order. The default is
+        `False`.
+
+    Returns
+    -------
+    out : list
+        The sorted input.
+
+    See Also
+    --------
+    natsorted
+    os_sort_keygen
+
+    Notes
+    -----
+    This will implicitly coerce all inputs to str before collating.
+
+    """
+    return sorted(seq, key=os_sort_keygen(key), reverse=reverse)
